import java.awt.*;
import java.awt.event.*;
import java.text.NumberFormat;

import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLCapabilities;
import javax.swing.*;
import javax.swing.text.NumberFormatter;

import com.sun.opengl.util.FPSAnimator;

/**
 * Initialization class for displaying a toroidal spiral and provides side controls.
 * @author Damian Johnson
 * @version November 10 2007
 */
public class Starter {
  private static final Dimension INITIAL_DISPLAY_SIZE = new Dimension(500, 500);
  
  public static void main(String[] args) {
    JFrame frame = new JFrame();
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.setTitle("Toroidal Spiral");
    frame.setResizable(false);
    
    GLCapabilities caps = new GLCapabilities();
    caps.setDoubleBuffered(true); // Enables double buffering
    caps.setHardwareAccelerated(true);
    
    GLCanvas panel = new GLCanvas(caps);
    ViewTracker tracker = new ViewTracker();
    // GLJPanel panel = new GLJPanel(caps); // Swing variant lacks hardware acceleration
    ToroidalRenderer renderer = new ToroidalRenderer(ToroidalRenderer.DrawType.SOLID, tracker);
    panel.addGLEventListener(renderer);
    panel.addMouseListener(tracker);
    panel.addMouseMotionListener(tracker);
    panel.setPreferredSize(INITIAL_DISPLAY_SIZE);
    
    final FPSAnimator animator = new FPSAnimator(panel, 20);
    animator.setRunAsFastAsPossible(false); // Restricts rate that display is called
    
    frame.addWindowListener(new WindowAdapter() {
      // Stops animator when closing
      public void windowClosing(WindowEvent e) {
        new Thread(new Runnable() {
          public void run() {
            animator.stop();
            System.exit(0);
          }
        }).start();
      }
    });
    animator.start();
    
    frame.add(panel);
    
    // Adds side controls
    JPanel wrapper = new JPanel(new BorderLayout());
    wrapper.add(makeControls(renderer), BorderLayout.NORTH);
    frame.add(wrapper, BorderLayout.EAST);
    
    frame.pack();
    frame.setVisible(true);
  }
  
  static JPanel makeControls(final ToroidalRenderer renderer) {
    JPanel controls = new JPanel(new GridLayout(0, 1));
    
    // Draw type combo box
    JPanel drawTypeLine = new JPanel(new BorderLayout());
    drawTypeLine.add(new JLabel(" Draw Type: "), BorderLayout.WEST);
    JComboBox drawTypes = new JComboBox();
    for (ToroidalRenderer.DrawType type : ToroidalRenderer.DrawType.values()) {
      String newItem = type.toString();
      drawTypes.addItem(newItem);
      if (type.equals(renderer.getDrawType())) drawTypes.setSelectedItem(newItem);
    }
    
    drawTypes.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JComboBox source = (JComboBox) event.getSource();
        renderer.setDrawType(ToroidalRenderer.DrawType.fromString((String) source.getSelectedItem()));
      }
    });
    drawTypeLine.add(drawTypes, BorderLayout.EAST);
    controls.add(drawTypeLine);
    
    // Materials combo box
    JPanel materialsLine = new JPanel(new BorderLayout());
    materialsLine.add(new JLabel(" Material: "), BorderLayout.WEST);
    JComboBox materials = new JComboBox();
    for (Material material : Material.values()) {
      String newItem = material.toString();
      materials.addItem(newItem);
      if (material.equals(renderer.getMaterial())) materials.setSelectedItem(newItem);
    }
    
    materials.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JComboBox source = (JComboBox) event.getSource();
        renderer.setMaterial(Material.fromString((String) source.getSelectedItem()));
      }
    });
    materialsLine.add(materials, BorderLayout.EAST);
    controls.add(materialsLine);
    
    // Twist trick check box
    JCheckBox segmented = new JCheckBox(" Segmented");
    segmented.setSelected(false);
    segmented.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JCheckBox source = (JCheckBox) event.getSource();
        renderer.setSegmentedSurface(source.isSelected());
      }
    });
    controls.add(segmented);
    
    // Twist trick check box
    JCheckBox twistTrick = new JCheckBox(" Twist Trick");
    twistTrick.setSelected(false);
    twistTrick.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JCheckBox source = (JCheckBox) event.getSource();
        renderer.setTwistTrick(source.isSelected());
      }
    });
    controls.add(twistTrick);
    
    // Rotation check box
    JCheckBox rotation = new JCheckBox(" Rotate");
    rotation.setSelected(true);
    rotation.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JCheckBox source = (JCheckBox) event.getSource();
        renderer.setRotationAnimation(source.isSelected());
      }
    });
    controls.add(rotation);
    
    // Divider
    controls.add(new JLabel());
    
    // Subdivision controls
    JPanel subdivisionsLine = new JPanel(new BorderLayout());
    subdivisionsLine.add(new JLabel(" Radius Subdivisions: "), BorderLayout.WEST);
    
    NumberFormatter tripleDigitFormat =
      new NumberFormatter(NumberFormat.getIntegerInstance());
    tripleDigitFormat.setMinimum(1);
    tripleDigitFormat.setMaximum(999);
    final JFormattedTextField subdivisions = new JFormattedTextField(tripleDigitFormat);
    subdivisions.setColumns(3);
    subdivisions.setValue(25);
    subdivisions.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        int value = ((Number) source.getValue()).intValue();
        renderer.setTorusApproximation(value, false);
      }
    });
    subdivisionsLine.add(subdivisions, BorderLayout.EAST);
    controls.add(subdivisionsLine);
    
    JCheckBox animateSubdivisions = new JCheckBox(" Animate Subdivisions");
    animateSubdivisions.setSelected(false);
    animateSubdivisions.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JCheckBox source = (JCheckBox) event.getSource();
        subdivisions.setEnabled(!source.isSelected());
        if (source.isSelected()) {
          renderer.setTorusApproximation(2, true);
        } else {
          int value = ((Number) subdivisions.getValue()).intValue();
          renderer.setTorusApproximation(value, false);
        }
      }
    });
    controls.add(animateSubdivisions);
    
    // Divider
    controls.add(new JLabel());
    
    // Controls for torus properties
    NumberFormatter doubleDigitFormat =
      new NumberFormatter(NumberFormat.getIntegerInstance());
    doubleDigitFormat.setMinimum(1);
    doubleDigitFormat.setMaximum(99);
    
    controls.add(new JLabel(" Torus/Spiral Properties:"));
    String indent = "     ";
    
    // Samplings field
    JPanel samplingsLine = new JPanel(new BorderLayout());
    samplingsLine.add(new JLabel(indent + "Samplings: "), BorderLayout.WEST);
    
    final JFormattedTextField samplingsField = new JFormattedTextField(tripleDigitFormat);
    samplingsField.setColumns(3);
    samplingsField.setValue(renderer.samplings);
    samplingsField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        renderer.samplings = ((Number) source.getValue()).intValue();
      }
    });
    samplingsLine.add(samplingsField, BorderLayout.EAST);
    controls.add(samplingsLine);
    
    // Inner radius field
    JPanel innerRadiusLine = new JPanel(new BorderLayout());
    innerRadiusLine.add(new JLabel(indent + "Inner Radius: "), BorderLayout.WEST);
    
    final JFormattedTextField innerRadiusField = new JFormattedTextField(tripleDigitFormat);
    innerRadiusField.setColumns(3);
    innerRadiusField.setValue(renderer.innerRadius);
    innerRadiusField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        renderer.innerRadius = ((Number) source.getValue()).intValue();
      }
    });
    innerRadiusLine.add(innerRadiusField, BorderLayout.EAST);
    controls.add(innerRadiusLine);
    
    // Outer radius field
    JPanel outerRadiusLine = new JPanel(new BorderLayout());
    outerRadiusLine.add(new JLabel(indent + "Outer Radius: "), BorderLayout.WEST);
    
    final JFormattedTextField outerRadiusField = new JFormattedTextField(doubleDigitFormat);
    outerRadiusField.setColumns(2);
    outerRadiusField.setValue(renderer.outerRadius);
    outerRadiusField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        renderer.outerRadius = ((Number) source.getValue()).intValue();
      }
    });
    outerRadiusLine.add(outerRadiusField, BorderLayout.EAST);
    controls.add(outerRadiusLine);
    
    // Torus radius field
    JPanel torusRadiusLine = new JPanel(new BorderLayout());
    torusRadiusLine.add(new JLabel(indent + "Torus Radius: "), BorderLayout.WEST);
    
    final JFormattedTextField torusRadiusField = new JFormattedTextField(doubleDigitFormat);
    torusRadiusField.setColumns(2);
    torusRadiusField.setValue(renderer.torusRadius);
    torusRadiusField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        renderer.torusRadius = ((Number) source.getValue()).intValue();
      }
    });
    torusRadiusLine.add(torusRadiusField, BorderLayout.EAST);
    controls.add(torusRadiusLine);
    
    // Inner coils field
    JPanel innerCoilsLine = new JPanel(new BorderLayout());
    innerCoilsLine.add(new JLabel(indent + "Inner Coils: "), BorderLayout.WEST);
    
    final JFormattedTextField innerCoilsField = new JFormattedTextField(doubleDigitFormat);
    innerCoilsField.setColumns(2);
    innerCoilsField.setValue(renderer.innerCoils);
    innerCoilsField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        renderer.innerCoils = ((Number) source.getValue()).intValue();
      }
    });
    innerCoilsLine.add(innerCoilsField, BorderLayout.EAST);
    controls.add(innerCoilsLine);
    
    // Wraps field
    JPanel wrapsLine = new JPanel(new BorderLayout());
    wrapsLine.add(new JLabel(indent + "Wraps: "), BorderLayout.WEST);
    
    final JFormattedTextField wrapsField = new JFormattedTextField(doubleDigitFormat);
    wrapsField.setColumns(2);
    wrapsField.setValue(renderer.outerCoils);
    wrapsField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JFormattedTextField source = (JFormattedTextField) event.getSource();
        renderer.outerCoils = ((Number) source.getValue()).intValue();
      }
    });
    wrapsLine.add(wrapsField, BorderLayout.EAST);
    controls.add(wrapsLine);
    
    return controls;
  }
}
