import java.awt.*;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;

import util.StringTools;

/**
 * Draws a spiral about a torus with several options for how its displayed.
 * @author Damian Johnson
 * @version November 10 2007
 */
public class ToroidalRenderer implements GLEventListener {
  private static final Color BACKGROUND = Color.BLACK;
  private static final float ROTATION_SPEED = 0.35f;
  private static final int TORUS_APPROXIMATION_RATE = 20; // The higher the slower slices subdivide
  private static final GLU glu = new GLU();
  private Material texture = Material.BLUE_PLASTIC;
  private DrawType drawType;
  private ViewTracker viewTracker; // Mouse listener for altering the view
  private boolean animateRotation = true;
  private int rotationCounter = 0;
  private boolean animateTorusApproximation = false;
  private int refreshCounter = 0; // Counter used in limiting rate of incrementing subdivisions
  private boolean useTwistTrick = false; // Performs a 75% rotation in edges used
  private boolean segmented = false; // Inverts the ordering that normals are used
  
  // Properties of the torus and spiral (forgoing getters/setters to avoid the boiler plate)
  protected int samplings = 200; // Number of subsections used to approximate torus
  protected double innerRadius = 100; // Center to the middle of the torus
  protected double outerRadius = 10; // Torus distance from the ring it wraps around
  protected double torusRadius = 10; // Radius of the torus
  protected int innerCoils = 10; // Number of turns around the inside of the torus
  protected int outerCoils = 1; // Number of wraps around the origin
  private int torusApproximation = 20; // Number of faces encircling each slice of the torus
  
  // Used for perspective projection
  private double near = 0.1;
  private double far = 2000.0;
  private double[] eye = {350.0, 0.0, 0.0};
  private double[] lookat = {0.0, 0.0, 0.0};
  private double[] up = {0.0, 0.0, 1.0};
  
  // Lighting
  private float[] lmodel_ambient    = {0.1f, 0.1f, 0.1f, 1.0f};
  
  private float[] light0_position   = {500.0f, 0.0f, 300.0f, 1.0f};
  private float[] light0_ambient    = {0.3f, 0.3f, 0.3f, 1.0f};
  private float[] light0_diffuse    = {1.0f, 1.0f, 1.0f, 1.0f};
  private float[] light0_specular   = {1.0f, 1.0f, 1.0f, 1.0f};
  
  public ToroidalRenderer(DrawType drawType, ViewTracker tracker) {
    this.drawType = drawType;
    this.viewTracker = tracker;
  }
  
  public void setDrawType(DrawType drawType) {
    this.drawType = drawType;
  }
  
  public DrawType getDrawType() {
    return this.drawType;
  }
  
  public void setMaterial(Material material) {
    this.texture = material;
  }
  
  public Material getMaterial() {
    return this.texture;
  }
  
  public void setRotationAnimation(boolean animate) {
    this.animateRotation = animate;
  }
  
  public void setTorusApproximation(int subdivisions, boolean animate) {
    this.torusApproximation = Math.max(1, subdivisions);
    this.animateTorusApproximation = animate;
    this.refreshCounter = 0;
  }
  
  public void setTwistTrick(boolean enable) {
    this.useTwistTrick = enable;
  }
  
  public void setSegmentedSurface(boolean enable) {
    this.segmented = enable;
  }
  
  public void init(GLAutoDrawable drawable) {
    final GL gl = drawable.getGL();
    
    // Sets the background
    float bgR = BACKGROUND.getRed() / 255;
    float bgG = BACKGROUND.getGreen() / 255;
    float bgB = BACKGROUND.getBlue() / 255;
    gl.glClearColor(bgR, bgG, bgB, 1.0f);
    
    gl.glClearDepth(1.0f);
    gl.glEnable(GL.GL_DEPTH_TEST);
    
    // Enables lighting
    gl.glEnable(GL.GL_LIGHTING);
    gl.glEnable(GL.GL_LIGHT0);
    
    gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, lmodel_ambient, 0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, light0_ambient, 0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, light0_diffuse, 0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, light0_specular, 0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, light0_position, 0);
    
    // Hidden face removal for efficiency
    gl.glEnable(GL.GL_CULL_FACE);
    gl.glCullFace(GL.GL_BACK);
  }
  
  public void display(GLAutoDrawable drawable) {
    ++this.refreshCounter;
    if (this.animateTorusApproximation && this.refreshCounter % TORUS_APPROXIMATION_RATE == 0) {
      ++this.torusApproximation;
    }
    
    final GL gl = drawable.getGL();
    
    gl.glMatrixMode(GL.GL_MODELVIEW);
    gl.glLoadIdentity();
    glu.gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], up[0], up[1], up[2]);
    
    
    gl.glClear(GL.GL_COLOR_BUFFER_BIT); // Clears the screen
    gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
    
    this.viewTracker.transformView(gl);
    if (this.animateRotation) ++this.rotationCounter;
    gl.glRotatef(ROTATION_SPEED * this.rotationCounter, 0f, 0f, 1f);
    
    // Applies texture
    this.texture.applyTexture(gl);
    
    if (this.drawType == DrawType.LINE) {
      gl.glBegin(GL.GL_LINE_LOOP);
      for (int i = 0; i < this.samplings; ++i) {
        double t = 2.0 * Math.PI * i / this.samplings;
        Point3D loc = getSpine(t, 0);
        gl.glVertex3d(loc.x, loc.y, loc.z);
      }
      gl.glEnd();
    } else {
      Point3D tube[][] = new Point3D[this.samplings][this.torusApproximation];
      Point3D tubeNormals[][] = new Point3D[this.samplings][this.torusApproximation];
      
      // Fills arrays with new point instances
      for (int i = 0; i < this.samplings; ++i) {
        for (int j = 0; j < this.torusApproximation; ++j) {
          tube[i][j] = new Point3D();
          tubeNormals[i][j] = new Point3D();
        }
      }
      
      double dt = 2 * Math.PI / this.samplings;
      double du = 2 * Math.PI / this.torusApproximation;
      double t = 0.0;
      for (int i = 0; i < this.samplings; ++i, t += dt) {
        Point3D c = getSpine(t, 0);
        Point3D dc = getSpine(t, 1);
        Point3D ddc = getSpine(t, 2);
        
        Point3D cb = getCrossProduct(dc, ddc);
        cb = getNormalized(cb);
        dc = getNormalized(dc);
        Point3D cn = getCrossProduct(cb, dc);
        
        double u = 0.0;
        for (int j = 0; j < this.torusApproximation; ++j, u += du) {
          tubeNormals[i][j].x = Math.cos(u) * cb.x + Math.sin(u) * cn.x;
          tubeNormals[i][j].y = Math.cos(u) * cb.y + Math.sin(u) * cn.y;
          tubeNormals[i][j].z = Math.cos(u) * cb.z + Math.sin(u) * cn.z;
          
          tube[i][j].x = c.x + this.torusRadius * tubeNormals[i][j].x;
          tube[i][j].y = c.y + this.torusRadius * tubeNormals[i][j].y;
          tube[i][j].z = c.z + this.torusRadius * tubeNormals[i][j].z;
        }
      }
      
      if (this.drawType == DrawType.WIRE_FRAME) {
        for (int i = 0; i < tube.length; ++i) {
          for (int j = 0; j < tube[i].length; ++j) {
            // Draws lines encircling torus
            int nextJ = (j + 1) % tube[i].length;
            gl.glBegin(GL.GL_LINES);
            gl.glVertex3d(tube[i][j].x, tube[i][j].y, tube[i][j].z);
            gl.glVertex3d(tube[i][nextJ].x, tube[i][nextJ].y, tube[i][nextJ].z);
            gl.glEnd();
            
            // Draws lines to next ring
            nextJ = this.useTwistTrick ? (j + (int) (.75 * this.torusApproximation)) % tube[i].length : j;
            int nextI = (i + 1) % tube.length;
            gl.glBegin(GL.GL_LINES);
            gl.glVertex3d(tube[i][j].x, tube[i][j].y, tube[i][j].z);
            gl.glVertex3d(tube[nextI][nextJ].x, tube[nextI][nextJ].y, tube[nextI][nextJ].z);
            gl.glEnd();
          }
        }
      } else if (this.drawType == DrawType.SOLID) {
        for (int i = 0; i < tube.length; ++i) {
          gl.glBegin(GL.GL_QUAD_STRIP);
          
          for (int jTmp = 0; jTmp < tube[i].length + 1; ++jTmp) {
            int j = jTmp % tube[i].length;
            
            int nextJ = this.useTwistTrick ? (j + (int) (.75 * this.torusApproximation)) % tube[i].length : j;
            int nextI = (i + 1) % tube.length;
            
            if (!this.segmented) gl.glNormal3d(tubeNormals[i][j].x, tubeNormals[i][j].y, tubeNormals[i][j].z);
            gl.glVertex3d(tube[i][j].x, tube[i][j].y, tube[i][j].z);
            if (this.segmented) gl.glNormal3d(tubeNormals[i][j].x, tubeNormals[i][j].y, tubeNormals[i][j].z);
            
            if (!this.segmented) gl.glNormal3d(tubeNormals[nextI][nextJ].x, tubeNormals[nextI][nextJ].y, tubeNormals[nextI][nextJ].z);
            gl.glVertex3d(tube[nextI][nextJ].x, tube[nextI][nextJ].y, tube[nextI][nextJ].z);
            if (this.segmented) gl.glNormal3d(tubeNormals[nextI][nextJ].x, tubeNormals[nextI][nextJ].y, tubeNormals[nextI][nextJ].z);
          }
          gl.glEnd();
        }
      } else {
        assert false : "Unrecognized draw type: " + this.drawType;
      }
    }
    
    gl.glFlush(); // The double buffering was handled by initialization capacities
  }
  
  public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {}
  
  public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
    final GL gl = drawable.getGL();
    gl.glViewport(0, 0, width, height);
    
    gl.glMatrixMode(GL.GL_PROJECTION);
    gl.glLoadIdentity();
    float aspect = width / Math.max(1, height);
    glu.gluPerspective(40.0, aspect, this.near, this.far);
    
    gl.glMatrixMode(GL.GL_MODELVIEW);
    gl.glLoadIdentity();
    glu.gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], up[0], up[1], up[2]);
  }
  
  /**
   * Provides a point along the toroidal spiral.
   * @param t sampling angle around the origin (between 0 and 2 PI)
   * @param derivatives number of derivative operations performed on the returned point- this method
   *        either provides the coordinates on the spine, first, or second derivatives for 0, 1, and
   *        2 respectively (other values not accepted)
   * @return location along the spiral with the proper number of derivative operations
   */
  private Point3D getSpine(double t, int derivatives) {
    if (t < 0 || t > Math.PI * 2) {
      String message = "getSpine: 't' must be between 0 and 2 PI (it was " + t + ")";
      throw new IllegalArgumentException(message);
    }
    
    if (derivatives < 0 || derivatives > 2) {
      String message =
          "getSpine: 'derivatives' must be between 0 and 2 (it was " + derivatives + ")";
      throw new IllegalArgumentException(message);
    }
    
    Point3D loc = new Point3D();
    if (derivatives == 0) {
      double temp = this.innerRadius + this.outerRadius * Math.cos(this.innerCoils * t);
      loc.x = temp * Math.cos(this.outerCoils * t);
      loc.y = temp * Math.sin(this.outerCoils * t);
      loc.z = this.outerRadius * Math.sin(this.innerCoils * t);
    } else if (derivatives == 1) {
      Point3D spineLoc = getSpine(t, 0);
      double temp = this.outerRadius * this.innerCoils * Math.sin(this.innerCoils * t);
      loc.x = -this.outerCoils * spineLoc.y - temp * Math.cos(this.outerCoils * t);
      loc.y = this.outerCoils * spineLoc.x - temp * Math.sin(this.outerCoils * t);
      loc.z = this.outerRadius * this.innerCoils * Math.cos(this.innerCoils * t);
    } else if (derivatives == 2) {
      Point3D firstDeriv = getSpine(t, 1);
      double temp1 = this.outerCoils * Math.sin(this.innerCoils * t);
      double temp2 = this.innerCoils * Math.cos(this.innerCoils * t);
      loc.x =
          -this.outerCoils * firstDeriv.y + this.outerRadius * this.innerCoils
              * (temp1 * Math.sin(this.outerCoils * t) - temp2 * Math.cos(this.outerCoils * t));
      loc.y =
          this.outerCoils * firstDeriv.x - this.outerRadius * this.innerCoils
              * (temp1 * Math.cos(this.outerCoils * t) + temp2 * Math.sin(this.outerCoils * t));
      loc.z = -1 * Math.pow(this.innerCoils, 2) * this.outerRadius * Math.sin(this.innerCoils * t);
    }
    
    return loc;
  }
  
  // Provides the cross product of two points representing vectors about the origin
  private Point3D getCrossProduct(Point3D p1, Point3D p2) {
    Point3D result = new Point3D();
    result.x = p1.y * p2.z - p1.z * p2.y;
    result.y = -(p2.z * p1.x - p2.x * p1.z);
    result.z = p1.x * p2.y - p1.y * p2.x;
    return result;
  }
  
  // Provides the point normalized about the origin
  private Point3D getNormalized(Point3D pt) {
    double length = Math.sqrt(Math.pow(pt.x, 2) + Math.pow(pt.y, 2) + Math.pow(pt.z, 2));
    
    Point3D result = new Point3D();
    result.x = pt.x / length;
    result.y = pt.y / length;
    result.z = pt.z / length;
    return result;
  }
  
  // Container for x, y, z coordinates with double precision
  private class Point3D {
    double x = 0;
    double y = 0;
    double z = 0;
    
    public Point3D() {}
    
    public Point3D(double x, double y, double z) {
      this.x = x;
      this.y = y;
      this.z = z;
    }
  }
  
  public enum DrawType {
    LINE, WIRE_FRAME, SOLID;
    
    /**
     * Returns the enum representation of a string. This is case sensitive.
     * @param str toString representation of this enum
     * @return enum associated with a string
     * @throws IllegalArgumentException if argument is not represented by this enum.
     */
    public static DrawType fromString(String str) {
      for (DrawType type : DrawType.values()) {
        if (str.equals(type.toString())) return type;
      }
      throw new IllegalArgumentException();
    }
    
    @Override
    public String toString() {
      return StringTools.getReadableConstant(this.name());
    }
  }
}
