import java.awt.*;
import java.awt.event.*;
import javax.swing.SwingUtilities;

import javax.media.opengl.GL;

/**
 * Captures mouse events and uses them to calculate a translation for viewing position and angle. By
 * default rotation is bound with the ctrl key, z axis alterations are bound with shift, and scaling
 * with alt.<br>
 * <br>
 * NOTE: Some alterations have been made to the controls to provide intuitive controls for the
 * ToroidalRenderer (which had redefined axis). This class should be revised before being reused
 * again.
 * @author Damian Johnson
 * @version February 10, 2007
 */
public class ViewTracker implements MouseListener, MouseMotionListener {
  private static final float SCALING_FACTOR = 0.01f;
  private Point pivot; // offset position for dragging, null when released
  private Tracker offsetTracker = new Tracker();
  private Tracker rotationTracker = new Tracker();
  private KeyBinding zAxisBinding = KeyBinding.SHIFT;
  private boolean enableZAlterations = true;
  private KeyBinding scaleBinding = KeyBinding.ALT;
  private boolean scalingEnabled = false;
  private float scale = 1f;
  
  public ViewTracker() {
    this.offsetTracker.binding = KeyBinding.CTRL;
  }
  
  public void mouseEntered(MouseEvent event) {}
  
  public void mouseExited(MouseEvent event) {}
  
  public void mouseClicked(MouseEvent event) {}
  
  public void mouseMoved(MouseEvent event) {}
  
  public void mousePressed(MouseEvent event) {
    if (SwingUtilities.isLeftMouseButton(event)) {
      this.pivot = event.getPoint();
    }
  }
  
  public void mouseReleased(MouseEvent event) {
    if (SwingUtilities.isLeftMouseButton(event)) {
      this.pivot = null;
    }
  }
  
  public void mouseDragged(MouseEvent event) {
    if (!SwingUtilities.isLeftMouseButton(event)) return;
    Point loc = event.getPoint();
    float dx = loc.x - this.pivot.x;
    float dy = loc.y - this.pivot.y;
    this.pivot = loc;
    
    if (this.offsetTracker.enabled) {
      /*
      if (this.enableZAlterations && this.offsetTracker.binding.isDown(event, this.zAxisBinding)
          && this.zAxisBinding.isDown(event)) {*/
      if (this.enableZAlterations && this.zAxisBinding.isDown(event)) {
        // this.offsetTracker.z += dy;
        this.offsetTracker.x -= dy;
      } else if (this.offsetTracker.binding.isDown(event)) {
        this.offsetTracker.y += dx;
        // this.offsetTracker.y -= dy;
        this.offsetTracker.z -= dy;
      }
    }
    
    if (this.rotationTracker.enabled) {
      /*
      if (this.enableZAlterations && this.rotationTracker.binding.isDown(event, this.zAxisBinding)
          && this.zAxisBinding.isDown(event)) {*/
      if (false) {
        this.rotationTracker.z += dy;
      } else if (this.rotationTracker.binding.isDown(event)) {
        //this.rotationTracker.x += dy;
        //this.rotationTracker.y -= dx;
        this.rotationTracker.y += dy;
        this.rotationTracker.z += dx;
      }
    }
    
    if (this.scalingEnabled && this.scaleBinding.isDown(event)) {
      this.scale += dy * SCALING_FACTOR;
    }
  }
  
  /**
   * Sets if mouse events can effect given offset values.
   * @param transformaiton type of offset
   * @param enabled if mouse events are registered
   */
  public void setEnabled(Transformation transformation, boolean enabled) {
    switch (transformation) {
      case OFFSET:
        this.offsetTracker.enabled = enabled;
      case ROTATION:
        this.rotationTracker.enabled = enabled;
      case SCALING:
        this.scalingEnabled = enabled;
    }
  }
  
  /**
   * Binds a key to a given type of offset.
   * @param transformaiton type of offset to bind
   * @param binding binding key
   */
  public void setBinding(Transformation transformation, KeyBinding binding) {
    switch (transformation) {
      case OFFSET:
        this.offsetTracker.binding = binding;
      case ROTATION:
        this.rotationTracker.binding = binding;
      case SCALING:
        this.scaleBinding = binding;
    }
  }
  
  /**
   * Binds a key to alterations on the z axis.
   * @param binding binding key
   */
  public void setZBinding(KeyBinding binding) {
    this.zAxisBinding = binding;
  }
  
  /**
   * Determines if alterations can be made to the z-axis.
   * @param enable if true z-axis changes can be made
   */
  public void enableZBinding(boolean enable) {
    this.enableZAlterations = enable;
  }
  
  /**
   * Resets offset and rotations values.
   */
  public void reset() {
    this.offsetTracker.x = 0f;
    this.offsetTracker.y = 0f;
    this.offsetTracker.z = 0f;
    
    this.rotationTracker.x = 0f;
    this.rotationTracker.y = 0f;
    this.rotationTracker.z = 0f;
    
    this.scale = 1f;
  }
  
  /**
   * Translates the view matrix for the current values held by this listener.
   * @param gl GL object to be altered
   */
  public void transformView(GL gl) {
    gl.glTranslatef(this.offsetTracker.x, this.offsetTracker.y, this.offsetTracker.z);
    gl.glScalef(this.scale, this.scale, this.scale);
    gl.glRotatef(this.rotationTracker.x, 1.0f, 0.0f, 0.0f);
    gl.glRotatef(this.rotationTracker.y, 0.0f, 1.0f, 0.0f);
    gl.glRotatef(this.rotationTracker.z, 0.0f, 0.0f, 1.0f);
    
    gl.glRotatef(20, 0.0f, 1.0f, 0.0f); // Offset for an initial position
  }
  
  private class Tracker {
    private float x = 0f;
    private float y = 0f;
    private float z = 0f;
    private boolean enabled = true;
    private KeyBinding binding = KeyBinding.NONE;
  }
  
  /**
   * Types of supported offsets.
   */
  public enum Transformation {
    OFFSET, ROTATION, SCALING;
  }
  
  /**
   * Key masks to be checked with mouse events.
   */
  public enum KeyBinding {
    NONE, SHIFT, ALT, CTRL;
    
    // Returns if this key is depressed in a mouse event
    private boolean isDown(MouseEvent event) {
      switch (this) {
        case NONE: {
          boolean anyDown = false;
          anyDown |= event.isShiftDown();
          anyDown |= event.isAltDown();
          anyDown |= event.isControlDown();
          return !anyDown;
        }
        case SHIFT:
          return event.isShiftDown();
        case ALT:
          return event.isAltDown();
        case CTRL:
          return event.isControlDown();
        default:
          return false;
      }
    }
    
    // Returns if this key is depressed without consideration for one key if none
    /*-
    private boolean isDown(MouseEvent event, KeyBinding exclude) {
      if (this == NONE) {
        boolean anyDown = false;
        for (KeyBinding binding : KeyBinding.values()) {
          if (binding == NONE || binding == exclude) continue;
          anyDown |= binding.isDown(event);
        }
        return !anyDown;
        
      } else {
        return isDown(event);
      }
    }
    */
  }
}
